import javax.microedition.midlet.*;
import javax.microedition.lcdui.*;
import java.util.*;

public class MetroNavigator extends MIDlet
{ 
    private static MetroNavigator instance = null;

    private Metro   metro;
    private Station from;
    private Station till;

    Command cmd;
    Route   minTimeRoute;
    Route   minCrossesRoute;

    static final Command QUIT_CMD = new Command(Locale.QUIT, Command.BACK, 5);
    static final Command BACK_CMD = new Command(Locale.BACK, Command.BACK, 1);
    static final Command START_CMD = new Command(Locale.START, Command.ITEM, 2);
    static final Command END_CMD = new Command(Locale.END, Command.ITEM, 3);
    static final Command SELECT_CMD = new Command(Locale.SELECT, Command.ITEM, 1);
    static final Command SEARCH_CMD = new Command(Locale.SEARCH, Command.ITEM, 2);
    static final Command MIN_TIME_CMD = new Command(Locale.MIN_TIME, Command.ITEM, 1);
    static final Command MIN_CROSSES_CMD = new Command(Locale.MIN_CROSSES, Command.ITEM, 1);
    static final Command SCROLL_CMD = new Command(Locale.SCROLL, Command.ITEM, 1);

    SchemeViewer viewer;

    public MetroNavigator() 
    {
        Scheme scheme = new Scheme("/scheme.xml");
        metro = new Metro(scheme.metroLines, scheme.metroCrosses);
        instance = this;
    }

    static MetroNavigator getInstance()
    {
        return instance;
    }
    
    public void setStation(Station station) 
    { 
        if (cmd == START_CMD) 
        { 
            from = station;
        } 
        else 
        {
            till = station;
        }
        Display.getDisplay(this).setCurrent(viewer);
        if (from != null && till != null) 
        { 
            minTimeRoute = metro.findRoute(from, till, 0);
            minCrossesRoute = metro.findRoute(from, till, 100);
            showRoute(minTimeRoute);
        }
    }

    protected void destroyApp(boolean unconditional) 
    {
    }    

    protected  void pauseApp() 
    {
    }

    protected void startApp() 
    {
        viewer = new SchemeViewer(this);
    }

    public void quit() 
    {
	destroyApp(true);
	notifyDestroyed();
	instance = null;
    }

    static final int ERROR_MSG_TIMEOUT = 2000;

    void showAlert(AlertType type, String title, String text, int timeout)
    {
        Alert currentAlert = new Alert(title);
        currentAlert.setTimeout(timeout);
        currentAlert.setString(text);
        currentAlert.setType(type);
        Display disp = Display.getDisplay(this);
        disp.setCurrent(currentAlert, viewer);
    }

    public void showError(String text)
    {
        showAlert(AlertType.ERROR, Locale.ERROR, text, ERROR_MSG_TIMEOUT);
    }

    void showRoute(Route route) 
    {
        String text = Locale.TOTAL_TIME + route.time
            + Locale.CROSSES + route.crosses.length
            + Locale.STATIONS + route.nStations;
        for (int i = 0; i < route.crosses.length; i++) 
        { 
            String from = route.crosses[i].src.getShortName();
            String till = route.crosses[i].dst.getShortName();
            if (from.equals(till)) 
            { 
                text += "\n" + (i+1) + Locale.CROSS_AT + from
                    + Locale.TO_LINE + route.crosses[i].dst.line.name;
            } 
            else 
            {
                text += "\n" + (i+1) + Locale.CROSS_AT + from
                    + Locale.TO_STATION + till + Locale.OF_LINE + route.crosses[i].dst.line.name;
            }
        } 
        viewer.showRoute(route);
        showAlert(AlertType.INFO, 
                  Locale.ROUTE, text, Alert.FOREVER);
    }

    void selectStation(int x, int y) { 
        Enumeration e = metro.stations.elements();
        int min = Integer.MAX_VALUE;
        Station nearest = null;
        while (e.hasMoreElements()) { 
            Station s = (Station)e.nextElement();
            int d = (s.x - x)*(s.x - x) + (s.y - y)*(s.y - y);
            if (d < min) { 
                min = d;
                nearest = s;
            }
        }
        if (nearest != null) { 
            if (from == null || till != null) { 
                from = nearest;
            } else { 
                till = nearest;
                minTimeRoute = metro.findRoute(from, till, 0);
                minCrossesRoute = metro.findRoute(from, till, 100);
                showRoute(minTimeRoute);
            }
        }
    }
}
