import javax.microedition.lcdui.*;
import java.util.*;

public class SchemeViewer extends Canvas implements CommandListener {
    MetroNavigator   navigator;
    static Image[][] scheme;
    private int      x, y;
    private int      width, height;
    private Route    route;
    private Timer    timer; 
    private int      tick;
    private int      currStation;
    private Image[]  imageCache;
    private int      imageCachePtr;
    private int[]    imageCacheId;
    private int      imageCacheSize;
    private Timer    repeatTimer = new Timer();
    private Repeater repeater;

    static final int STEP = 50;

    static final int REPEAT_DELAY = 1000;
    static final int REPEAT_PERIOD = 100;

    static final int BLACK = 0;
    static final int WHITE = 0xffffff;
    static final int FREQ = 1000;

    private void setWindow(int stationNo) { 
        x = route.stations[stationNo].x - width/2;
        y = route.stations[stationNo].y - height/2;
        checkCoordinates();
    }

    SchemeViewer(MetroNavigator navigator) {
        this.navigator = navigator;
        width = getWidth();
        height = getHeight();

        imageCacheSize = 0;
        
        setCommandListener(this);
        addCommand(MetroNavigator.START_CMD);
        addCommand(MetroNavigator.END_CMD);
        addCommand(MetroNavigator.QUIT_CMD);
        x = width < Scheme.MAP_WIDTH ? (Scheme.MAP_WIDTH-width)/2/Scheme.MAP_CELL * Scheme.MAP_CELL : 0;
        y = height < Scheme.MAP_HEIGHT ? (Scheme.MAP_HEIGHT-height)/2/Scheme.MAP_CELL * Scheme.MAP_CELL : 0;
        Display.getDisplay(navigator).setCurrent(this);
    }


    void showRoute(Route route) { 
        this.route = route;
        if (route == navigator.minCrossesRoute) { 
            removeCommand(MetroNavigator.MIN_CROSSES_CMD);
            if (navigator.minTimeRoute != null) { 
                addCommand(MetroNavigator.MIN_TIME_CMD);
            }
        } else { 
            removeCommand(MetroNavigator.MIN_TIME_CMD);
            if (navigator.minCrossesRoute != null) { 
                addCommand(MetroNavigator.MIN_CROSSES_CMD);
            }
        }
        currStation = -1;
        addCommand(MetroNavigator.SCROLL_CMD);
        setWindow(0);
        timer = new Timer();
        timer.scheduleAtFixedRate(new RedrawTask(), FREQ, FREQ);
        repaint();
    }

    class RedrawTask extends TimerTask { 
        public void run() { 
            repaint();
        }
    }

    private Image getImage(int i) { 
        try { 
            return Image.createImage("/scheme" + i + ".png");
        } catch (Exception x) {
            return null;
        }
        /*
        try { 
            for (int k = 0; k < imageCacheSize; k++) { 
                if (imageCacheId[k] == i) { 
                    return imageCache[k];
                }
            }
            Image image;
            if (imageCacheSize == 0) { 
                Runtime rt = Runtime.getRuntime();                
                long freeMemoryBefore = rt.freeMemory();
                //System.out.println("free memory before" + freeMemoryBefore);
                image = Image.createImage("/scheme" + i + ".png");
                long freeMemoryAfter = rt.freeMemory();
                //System.out.println("free memory after" + freeMemoryAfter);
                int imageSize = (freeMemoryBefore == freeMemoryAfter) 
                    ? Scheme.MAP_WIDTH/Scheme.MAP_CELL*Scheme.MAP_HEIGHT/Scheme.MAP_CELL 
                    : (int)(freeMemoryBefore - freeMemoryAfter);
                int n = (int)(freeMemoryAfter / imageSize);
                if (n <= 0) { 
                    n = 1;
                } else if (n > Scheme.MAP_WIDTH/Scheme.MAP_CELL*Scheme.MAP_HEIGHT/Scheme.MAP_CELL) { 
                    n = Scheme.MAP_WIDTH/Scheme.MAP_CELL*Scheme.MAP_HEIGHT/Scheme.MAP_CELL;
                } else { 
                    n = 4;
                }
                //System.out.println("Image cache size " + n);
                imageCacheSize = n;
                imageCache = new Image[n];
                imageCacheId = new int[n];
                while (--n >= 0) { 
                    imageCacheId[n] = -1;
                }
            } else { 
                imageCache[imageCachePtr] = null;
                image = Image.createImage("/scheme" + i + ".png");
            }
            imageCache[imageCachePtr] = image;
            imageCacheId[imageCachePtr] = i;
            imageCachePtr = (imageCachePtr + 1) & (imageCacheSize-1);
            return image;
        } catch (Exception x) {
            System.err.println(x);
            return null;
        }
        */
    }

    protected void paint(Graphics g) {
        g.setColor(WHITE);
        g.fillRect(g.getClipX(), g.getClipY(), g.getClipWidth(), g.getClipHeight());
        for (int i = y/Scheme.MAP_CELL, y0 = i*Scheme.MAP_CELL; y0 < y + height && y0 < Scheme.MAP_HEIGHT; i++, y0 += Scheme.MAP_CELL) {
            for (int j = x/Scheme.MAP_CELL, x0 = j*Scheme.MAP_CELL; x0 < x + width && x0 < Scheme.MAP_WIDTH; j++, x0 += Scheme.MAP_CELL) {
                Image image = getImage(i*Scheme.MAP_WIDTH/Scheme.MAP_CELL + j + 1); 
                if (image != null) { 
                    g.drawImage(image, x0-x, y0-y, Graphics.TOP|Graphics.LEFT);
                } else { 
                    g.fillRect(x0-x, y0-y, Scheme.MAP_CELL, Scheme.MAP_CELL);
                }
            } 
        }
        if (route != null) {
            if ((++tick & 1) != 0) { 
                g.setColor(BLACK);
            } else { 
                g.setColor(WHITE);
            }
            g.setStrokeStyle(Graphics.DOTTED);
            for (int i = 1; i < route.stations.length; i++) { 
                int x0 = route.stations[i-1].x;
                int y0 = route.stations[i-1].y;
                int x1 = route.stations[i].x;
                int y1 = route.stations[i].y;
                if (x0 < x) { 
                    if (x1 < x) { 
                        continue;
                    }
                    y0 += (y1-y0)*(x-x0)/(x1-x0);
                    x0 = x;
                } else if (x0 > x + width) { 
                    if (x1 > x + width) {
                        continue;
                    }
                    y0 += (y1-y0)*(x+width-x0)/(x1-x0);
                    x0 = x + width;
                }
                if (x1 < x) { 
                    y1 += (y0-y1)*(x-x1)/(x0-x1);
                    x1 = x;
                } else if (x1 > x + width) { 
                    y1 += (y0-y1)*(x+width-x1)/(x0-x1);
                    x1 = x + width;
                }
                if (y0 < y) { 
                    if (y1 < y) { 
                        continue;
                    }
                    x0 += (x1-x0)*(y-y0)/(y1-y0);
                    y0 = y;
                } else if (y0 > y + height) { 
                    if (y1 > y + height) {
                        continue;
                    }
                    x0 += (x1-x0)*(y+height-y0)/(y1-y0);
                    y0 = y + height;
                }
                if (y1 < y) { 
                    x1 += (x0-x1)*(y-y1)/(y0-y1);
                    y1 = y;
                } else if (y1 > y + height) { 
                    x1 += (x0-x1)*(y+height-y1)/(y0-y1);
                    y1 = y + height;
                }                   
                g.drawLine(x0-x, y0-y, x1-x, y1-y);
            }
            if (currStation >= 0) { 
                g.fillArc(route.stations[currStation].x-x-2, 
                          route.stations[currStation].y-y-2, 5, 5,
                          0, 360);
                if ((tick & 1) == 0) { 
                    if (++currStation >= route.stations.length) { 
                        currStation = -1;
                    } else { 
                        setWindow(currStation);
                    }
                }
            } else { 
                for (int i = 0; i < route.stations.length; i++) { 
                    int x0 = route.stations[i].x;
                    int y0 = route.stations[i].y;
                    if (x0 >= x && y0 >= y && x0 < x + width && y0 < y + width) { 
                        g.fillArc(x0-x-2, y0-y-2, 5, 5, 0, 360);
                    }
                }
            }                    
        }
    }                    
                    

    public void commandAction(Command c, Displayable d)  
    {
        if (c == MetroNavigator.QUIT_CMD) {
            if (timer != null) { 
                timer.cancel();
            }
            navigator.quit();
        } else if (c == MetroNavigator.START_CMD) {
            navigator.cmd = c;
            new SelectStation();
        } else if (c == MetroNavigator.END_CMD) {
            navigator.cmd = c;
            new SelectStation();
        } else if (c == MetroNavigator.SCROLL_CMD) {
            currStation = 0;
            tick = 0;
            setWindow(0);
        } else if (c == MetroNavigator.MIN_CROSSES_CMD) { 
            navigator.showRoute(navigator.minCrossesRoute);
        } else if (c == MetroNavigator.MIN_TIME_CMD) { 
            navigator.showRoute(navigator.minTimeRoute);
        }
    }

    private void checkCoordinates() { 
        if (x < 0) x = 0;
        else if (x + width > Scheme.MAP_WIDTH) x = Scheme.MAP_WIDTH - width;
        if (y < 0) y = 0;
        else if (y + height > Scheme.MAP_HEIGHT) y = Scheme.MAP_HEIGHT - height;
    }

    class Repeater extends TimerTask { 
        Repeater(int keyCode) { 
            this.keyCode = keyCode;
        }

        public void run() { 
            keyAction(keyCode);
        }

        int keyCode;
    }

    public void pointerPressed(int px, int py) { 
        navigator.selectStation(x + px, y + py);
    }

    public void keyPressed(int keyCode) {
        if (repeater != null) { 
            repeater.cancel();
        }
        repeater = new Repeater(keyCode);
        repeatTimer.schedule(repeater, REPEAT_DELAY, REPEAT_PERIOD);
    }

    public void keyReleased(int keyCode) {
        keyAction(keyCode);
        if (repeater != null) { 
            repeater.cancel();
	    repeater = null;
        }
    }
    
    public void keyAction(int keyCode) {
	switch (keyCode) {
        case KEY_NUM1:
            x -= STEP;
            y -= STEP;
            break;
        case KEY_NUM2:
            y -= STEP;
            break;
        case KEY_NUM3:
            x += STEP;
            y -= STEP;
            break;
        case KEY_NUM4:
            x -= STEP;
            break;
        case KEY_NUM6:
            x += STEP;
            break;
        case KEY_NUM7:
            x -= STEP;
            y += STEP;
            break;
        case KEY_NUM8:
            y += STEP;
            break;
        case KEY_NUM9:
            x += STEP;
            y += STEP;
            break;
        default:
            switch (getGameAction(keyCode)) { 
            case LEFT:
                x -= STEP;
                break;
            case RIGHT:
                x += STEP;
                break;
            case UP:
                y -= STEP;
                break;
            case DOWN:
                y += STEP;
                break;
            default:
                return;
            }
        }
        currStation = -1;
        checkCoordinates();
        repaint();
    }
}
            
